'[--------------------------------------------------------------]
'
'   NAME: CSampleGraphicsEngine
'   AUTHOR: Jack Hoxley
'   CONTACT: mailto:Jack.Hoxley@DirectX4VB.com
'            http://www.DirectX4VB.com
'
'   DATE: 16th December 2002 (started)
'         21st December 2002 (finished)
'
'   NOTES:
'       This is a relatively simple graphics library designed
'       for learning from - it'll work fine for small-medium
'       projects, but it's far from a complete solution.
'
'       You're free to use/disect this class for your own
'       projects, although a little credit for the original
'       work wouldn't hurt!
'
'[--------------------------------------------------------------]

'these two imports bring all of the classes, types and
'objects that we need to write our application.
Imports Microsoft.DirectX
Imports Microsoft.DirectX.Direct3D

'a useful import for any multimedia-based application
Imports System.Math

'begin the class
Public Class CSampleGraphicsEngine

    'core objects
    Private D3DRoot As Manager
    Private D3DDev As Device
    Private D3DHelp As D3DX

    'transformation matrices
    Private matCube1 As Matrix
    Private matCube2 As Matrix
    Private matView As Matrix
    Private matProj As Matrix

    'textures
    Private texCube1 As Texture
    Private texCube2 As Texture
    Private texMenu As Texture

    'geometry
    Private vbCube As VertexBuffer
    Private vbMenu As VertexBuffer

    'fonts
    Private fntOut As Font

    'per-frame variables
    Private lastFrameUpdate As Int32
    Private cube1Angle As Single
    Private cube2Angle As Single
    Private Const cube1Speed As Single = 50.0F
    Private Const cube2Speed As Single = 75.0F
    Private Const cube1Size As Single = 2.0F
    Private Const cube2Size As Single = 4.0F

    'misc variables
    Private bInitOkay As Boolean = False
    Private iLastFPSCheck As Int32
    Private Const iFPSProfileSpeed As Integer = 200 'ms
    Private iCurrCnt As Integer
    Private iFrameRate As Integer
    Private sDevInfo As String
    Private sDispInfo As String

    'control variables:
    Private bRenderWireframe As Boolean = False
    Private bRenderTextures As Boolean = True
    Private bBackFaceCulling As Boolean = True
    Private bDrawFPS As Boolean = True
    Private bShowFrameRate As Boolean = True

    'reference variables:
    Private rTarget As System.Windows.Forms.Form
    Private bWindowed As Boolean
    Private iFontSize As Integer



    '[--------------------------------------------------------------]
    '   CONSTRUCTOR #1
    '   Creates a windowed application based on the current
    '   display properties and that of the Target specified
    '[--------------------------------------------------------------]
    Public Sub New(ByVal Target As Form)
        Try
            Dim d3dPP As New PresentParameters()

            d3dPP.Windowed = True : bWindowed = True
            d3dPP.SwapEffect = SwapEffect.Discard
            d3dPP.BackBufferCount = 1
            d3dPP.BackBufferFormat = D3DRoot.Adapters(0).CurrentDisplayMode.Format
            d3dPP.BackBufferWidth = Target.ClientSize.Width()
            d3dPP.BackBufferHeight = Target.ClientSize.Height()

            sDispInfo = "[WINDOWED] " + _
                        Target.ClientSize.Width.ToString + "x" + _
                        Target.ClientSize.Height.ToString + " " + d3dPP.BackBufferFormat.ToString()

            rTarget = Target

            initialiseDevice(CType(Target, Control), d3dPP)
        Catch err As Exception
            bInitOkay = False
            'you can examine 'err' and/or output that error, but I don't want
            'the calling application to get much info about the internal state of the engine.
            'for now, we'll just throw an exception that'll fall back to the caller
            Throw New Exception("Could not initialise graphics engine.")
        End Try
    End Sub

    '[--------------------------------------------------------------]
    '   CONSTRUCTOR #2
    '   Creates a fullscreen application based on the display properties
    '   specified. Throws an exception if the parameters are not valid.
    '[--------------------------------------------------------------]
    Public Sub New(ByVal Target As System.Windows.Forms.Form, _
                   ByVal iWidth As Integer, _
                   ByVal iHeight As Integer, _
                   ByVal iDepth As Integer _
                   )
        Try
            Dim d3dPP As New PresentParameters()
            If Not isDisplayModeOkay(iWidth, iHeight, iDepth) Then
                Throw New Exception("Could not verify that the selected " _
                                    + "resolution is compatable with the hardware")
            End If
            bWindowed = False
            d3dPP.BackBufferWidth = iWidth
            d3dPP.BackBufferHeight = iHeight
            d3dPP.BackBufferCount = 1
            d3dPP.SwapEffect = SwapEffect.Copy
            d3dPP.PresentationInterval = PresentInterval.Immediate
            Select Case iDepth
                Case 16
                    d3dPP.BackBufferFormat = Format.R5G6B5
                    sDispInfo = "[Fullscreen] " + iWidth.ToString + "x" + iHeight.ToString + " 16bit R5G6B5"
                Case 32
                    d3dPP.BackBufferFormat = Format.X8R8G8B8
                    sDispInfo = "[Fullscreen] " + iWidth.ToString + "x" + iHeight.ToString + " 32bit X8R8G8B8"
                Case Else
                    Throw New Exception("iDepth does not contain a valid value, only 16 or 32 is acceptable")
            End Select

            rTarget = Target

            initialiseDevice(CType(Target, System.Windows.Forms.Control), d3dPP)
        Catch err As Exception
            bInitOkay = False
            'you can examine 'err' and/or output that error, but I don't want
            'the calling application to get much info about the internal state of the engine.
            'for now, we'll just throw an exception that'll fall back to the caller
            Throw New Exception("Could not initialise graphics engine")
        End Try
    End Sub

    '[--------------------------------------------------------------]
    '   DESTRUCTOR #1
    '   Makes sure that all objects that were used are terminated if
    '   necessary.
    '[--------------------------------------------------------------]
    Protected Overrides Sub Finalize()

    End Sub


    '[--------------------------------------------------------------]
    '   isDisplayModeOkay()
    '   Checks the specified display mode parameters against
    '   the hardware to see if the requested mode is acceptable
    '[--------------------------------------------------------------]
    Public Shared Function isDisplayModeOkay(ByVal iWidth As Integer, _
                                      ByVal iHeight As Integer, _
                                      ByVal iDepth As Integer _
                                      ) As Boolean
        Try
            'this covers the most commonly used display modes
            'only. others exist and are used for more specific cases
            Dim AdapterInfo As AdapterInformation
            Dim DispMode As DisplayMode
            Dim fmt As Format
            Dim D3Dr As Manager

            Select Case iDepth
                Case 16
                    fmt = Format.R5G6B5
                Case 32
                    fmt = Format.X8R8G8B8
            End Select

            For Each AdapterInfo In D3Dr.Adapters
                For Each DispMode In AdapterInfo.SupportedDisplayModes(fmt)
                    If DispMode.Width = iWidth Then
                        If DispMode.Height = iHeight Then
                            Return True
                        End If
                    End If
                Next
            Next

            Throw New Exception("No compatable resolution was found")

            'errors are caught silently, no real need
            'to inform the user as to whether the resolution
            'is supported or not. The caller can do that.
        Catch DXErr As DirectXException
            Return False
        Catch Err As Exception
            Return False
        End Try
    End Function


    '[--------------------------------------------------------------]
    '   oneFrameUpdate()
    '   this updates all the math, physics and any other
    '   per-frame calculations necessary.
    '[--------------------------------------------------------------]
    Public Sub oneFrameUpdate()
        If Not bInitOkay Then Throw New Exception("cannot update frame: engine was not properly initialised")

        'general variables
        Dim matTmp As Matrix

        'calculate the current rotation angles for the two cubes
        cube1Angle += ((Environment.TickCount() - lastFrameUpdate) / 1000) * cube1Speed
        cube2Angle += ((Environment.TickCount() - lastFrameUpdate) / 1000) * cube2Speed
        lastFrameUpdate = Environment.TickCount()

        'calculate cube1's matrix
        matCube1 = Matrix.Identity()
        matCube1 = Matrix.Multiply(matCube1, _
                                   Matrix.Scaling(3, 3, 3))
        matCube1 = Matrix.Multiply(matCube1, _
                                   Matrix.RotationX(cube1Angle * (Math.PI / 180)))
        matCube1 = Matrix.Multiply(matCube1, _
                                   Matrix.RotationY(cube1Angle * (Math.PI / 180)))

        'calculate cube2's matrix
        matCube2 = Matrix.Identity()
        matCube2 = Matrix.Multiply(matCube2, Matrix.Scaling(3, 3, 3))
        matCube2 = Matrix.Multiply(matCube2, _
                   Matrix.RotationX(cube2Angle * (Math.PI / 180)))
        matCube2 = Matrix.Multiply(matCube2, Matrix.Translation(0, -6, 4))
        matCube2 = Matrix.Multiply(matCube2, Matrix.RotationZ(cube2Angle * (Math.PI / 180)))

        'calculate the current framerate
        If (Environment.TickCount() - iLastFPSCheck >= iFPSProfileSpeed) Then
            iLastFPSCheck = Environment.TickCount()
            iFrameRate = iCurrCnt * (1000 / iFPSProfileSpeed)
            iCurrCnt = 0
        End If
        iCurrCnt += 1
    End Sub

    '[--------------------------------------------------------------]
    '   oneFrameRender()
    '   this handles rendering all of the graphics. it
    '   shouldn't have to deal with updating any variables
    '   or objects.
    '[--------------------------------------------------------------]
    Public Sub oneFrameRender()
        If Not bInitOkay Then Throw _
            New Exception("cannot render frame: engine was not properly initialised")

        '[------------------------------------]
        '   0. Any variables
        '[------------------------------------]

        '[------------------------------------]
        '   1. clear the render-target
        '[------------------------------------]
        D3DDev.Clear(ClearFlags.Target Or ClearFlags.ZBuffer, _
                     Drawing.Color.FromArgb(255, 0, 0, 64), 1.0F, 0)

        '[------------------------------------]
        '   2. begin rendering
        '[------------------------------------]
        D3DDev.BeginScene()

        '[------------------------------------]
        '   3. render 3D objects
        '[------------------------------------]
        D3DDev.SetStreamSource(0, vbCube, 0)
        D3DDev.VertexFormat = CustomVertex.PositionTextured.Format

        'render the first cube
        If bRenderTextures Then
            D3DDev.SetTexture(0, texCube1)
        Else
            D3DDev.SetTexture(0, Nothing)
        End If
        D3DDev.Transform.World = matCube1
        D3DDev.DrawPrimitives(PrimitiveType.TriangleList, 0, 12)

        'render the second cube
        If bRenderTextures Then
            D3DDev.SetTexture(0, texCube2)
        Else
            D3DDev.SetTexture(0, Nothing)
        End If
        D3DDev.Transform.World = matCube2
        D3DDev.DrawPrimitives(PrimitiveType.TriangleList, 0, 12)

        '[------------------------------------]
        '   4. render 2D objects and text
        '[------------------------------------]

        'render the menu background
        Dim bPrev As Boolean = wireframe()
        wireframe = False
        D3DDev.RenderState.AlphaBlendEnable = True

        D3DDev.SetStreamSource(0, vbMenu, 0)
        D3DDev.VertexFormat = CustomVertex.TransformedTextured.Format
        D3DDev.SetTexture(0, texMenu)
        D3DDev.DrawPrimitives(PrimitiveType.TriangleStrip, 0, 2)

        D3DDev.RenderState.AlphaBlendEnable = False
        wireframe = bPrev

        'draw the important text
        If bDrawFPS Then
            fntOut.DrawText(sDevInfo, _
                            New Drawing.Rectangle(5, 5, 0, 0), _
                            DrawTextFormat.Left Or DrawTextFormat.Top, _
                            Drawing.Color.FromArgb(255, 200, 128, 64))
            fntOut.DrawText(sDispInfo, _
                            New Drawing.Rectangle(5, 5 + (iFontSize + 6), 0, 0), _
                            DrawTextFormat.Left Or DrawTextFormat.Top, _
                            Drawing.Color.FromArgb(255, 200, 128, 64))
            fntOut.DrawText("Frame Rate: " + iFrameRate.ToString + "fps", _
                            New Drawing.Rectangle(5, 5 + (2 * (iFontSize + 6)), 0, 0), _
                            DrawTextFormat.Left Or DrawTextFormat.Top, _
                            Drawing.Color.FromArgb(255, 200, 128, 64))
        End If

        'draw the 'menu' bar
        fntOut.DrawText(IIf(bRenderWireframe, "F1 - Solid", "F1 - Wireframe"), _
                        New Drawing.Rectangle(5, 5 + (5 * (iFontSize + 6)), 0, 0), _
                        DrawTextFormat.Left Or DrawTextFormat.Top, _
                        Drawing.Color.FromArgb(255, 200, 128, 64))
        fntOut.DrawText(IIf(bRenderTextures, "F2 - Shaded", "F2 - Textured"), _
                        New Drawing.Rectangle(5, 5 + (6 * (iFontSize + 6)), 0, 0), _
                        DrawTextFormat.Left Or DrawTextFormat.Top, _
                        Drawing.Color.FromArgb(255, 200, 128, 64))
        fntOut.DrawText(IIf(bBackFaceCulling, "F3 - No Culling", "F3 - Cull Back Faces"), _
                        New Drawing.Rectangle(5, 5 + (7 * (iFontSize + 6)), 0, 0), _
                        DrawTextFormat.Left Or DrawTextFormat.Top, _
                        Drawing.Color.FromArgb(255, 200, 128, 64))
        fntOut.DrawText(IIf(bDrawFPS, "F4 - Hide Status", "F4 - Show Status"), _
                        New Drawing.Rectangle(5, 5 + (8 * (iFontSize + 6)), 0, 0), _
                        DrawTextFormat.Left Or DrawTextFormat.Top, _
                        Drawing.Color.FromArgb(255, 200, 128, 64))
        fntOut.DrawText("ESC - quit demo", _
                        New Drawing.Rectangle(5, 5 + (9 * (iFontSize + 6)), 0, 0), _
                        DrawTextFormat.Left Or DrawTextFormat.Top, _
                        Drawing.Color.FromArgb(255, 200, 128, 64))



        '[------------------------------------]
        '   5. end rendering
        '[------------------------------------]
        D3DDev.EndScene()

        '[------------------------------------]
        '   6. present frame on the screen
        '[------------------------------------]
        D3DDev.Present()
    End Sub

    '[--------------------------------------------------------------]
    '   initialiseDevice()
    '   once the constructor has filled out the PresentParameters
    '   we can use this method to finish off all device
    '   initialisation
    '[--------------------------------------------------------------]
    Private Sub initialiseDevice(ByVal Target As System.Windows.Forms.Control, _
                                 ByVal win As PresentParameters)
        Try
            '0. declare useful variables:
            Dim D3DCaps As Caps
            Dim DevCreate As Integer

            '1. check for, and specify, depth buffer
            If D3DRoot.CheckDepthStencilMatch(0, DeviceType.Hardware, _
                                              win.BackBufferFormat, _
                                              win.BackBufferFormat, _
                                              DepthFormat.D16) Then
                'support exists.
                win.AutoDepthStencilFormat = DepthFormat.D16
                win.EnableAutoDepthStencil = True
            Else
                'support does not exist!
                Throw New Exception("Direct3D Adapter does not support 16bit depth buffers. This sample requires a depth buffer with at least 16bits of precision")
            End If

            '2. examine the device capabilities
            D3DCaps = D3DRoot.GetDeviceCaps(0, DeviceType.Hardware)


            'allow this engine to take advantage of hw tnl where available
            If D3DCaps.DeviceCaps.SupportsHardwareTransformAndLight Then
                DevCreate = CreateFlags.HardwareVertexProcessing Or CreateFlags.MultiThreaded
            Else
                DevCreate = CreateFlags.SoftwareVertexProcessing Or CreateFlags.MultiThreaded
            End If

            '3. attempt to create the device interface.
            D3DDev = New Device(0, DeviceType.Hardware, Target, DevCreate, win)
            If D3DDev Is Nothing Then Throw New Exception("device was not allocated")


            '4. Configure render states and other parameters
            'we need to tell D3D we want it to use the depth buffer
            D3DDev.RenderState.ZBufferEnable = True

            'we don't want to use lighting in this sample
            D3DDev.RenderState.Lighting = False

            'allow texture transparencies
            D3DDev.RenderState.SourceBlend = Blend.SourceAlpha
            D3DDev.RenderState.DestinationBlend = Blend.InvSourceAlpha

            'set up texture blending
            D3DDev.SamplerState(0).MinFilter = TextureFilter.Linear
            D3DDev.SamplerState(0).MagFilter = TextureFilter.Linear

            '5. load textures.
            loadTextures(win.BackBufferFormat)

            '6. create the necessary geometry.
            loadGeometry()

            '7. Setup the matrices.
            'view matrix: describes the properties of the camera. 
            D3DDev.Transform.View = Matrix.LookAtLH(New Vector3(0, 0, -15), _
                                                    New Vector3(0, 0, 0), _
                                                    New Vector3(0, 1, 0))

            'projection matrix: describes the properties of the cameras lens.
            D3DDev.Transform.Projection = Matrix.PerspectiveFovLH(Math.PI / 4, _
                                                                  4 / 3, _
                                                                  1, _
                                                                  100)

            'world matrix: how all the geometry is altered in world space.
            D3DDev.Transform.World = Matrix.Identity()

            '8. sort out fonts for on-screen rendering
            iFontSize = 11
            fntOut = New Font(D3DDev, New Drawing.Font("Arial", iFontSize, FontStyle.Bold))

            '9. Condigure any misc. or general variables

            'if we dont do this then the initial angle calculations
            'go pear shaped.
            lastFrameUpdate = Environment.TickCount()

            'store the name of the graphics card as the driver reports it
            sDevInfo = D3DRoot.Adapters(0).Information.Description

            'if we've got this far then we can assume it was a successful
            'initialisation.
            bInitOkay = True

        Catch err As Exception
            bInitOkay = False
            'we wont output the real error message, instead we'll just inform the
            'calling object that we failed - and let them handle it. It would be worth
            'investing time in a developing a more detailed error-handling system.
            Throw New Exception("Could not initialise graphics engine.")
        End Try
    End Sub


    '[--------------------------------------------------------------]
    '   loadTextures()
    '   this loads all the required textures from disk
    '   into memory
    '[--------------------------------------------------------------]
    Private Sub loadTextures(ByVal AdapterFmt As Format)
        'attempt to load the menu texture
        If D3DRoot.CheckDeviceFormat(0, _
                                     DeviceType.Hardware, _
                                     AdapterFmt, _
                                     0, _
                                     ResourceType.Textures, _
                                     Format.A8R8G8B8 _
                                     ) Then
            texMenu = TextureLoader.FromFile(D3DDev, _
                                             Application.StartupPath + "\menubar.bmp", _
                                             256, 128, 1, 0, _
                                             Format.A8R8G8B8, Pool.Managed, _
                                             Filter.None, Filter.None, _
                                             Drawing.Color.Magenta.ToArgb())

        ElseIf D3DRoot.CheckDeviceFormat(0, _
                                         DeviceType.Hardware, _
                                         AdapterFmt, _
                                         0, _
                                         ResourceType.Textures, _
                                         Format.A1R5G5B5 _
                                         ) Then
            texMenu = TextureLoader.FromFile(D3DDev, _
                                             Application.StartupPath + "\menubar.bmp", _
                                             256, 128, 1, 0, _
                                             Format.A1R5G5B5, Pool.Managed, _
                                             Filter.None, Filter.None, _
                                             Drawing.Color.Magenta.ToArgb())

        Else
            Throw New Exception("texMenu cannot be created: no support for 16 bit or 32bit textures with at least 1bit of alpha")
        End If

        'attempt to load the cube textures
        If D3DRoot.CheckDeviceFormat(0, _
                                     DeviceType.Hardware, _
                                     AdapterFmt, _
                                     0, _
                                     ResourceType.Textures, _
                                     Format.X8R8G8B8 _
                                     ) Then
            texCube1 = TextureLoader.FromFile(D3DDev, _
                                              Application.StartupPath + "\cube1.bmp", _
                                              256, 256, 1, 0, _
                                              Format.X8R8G8B8, Pool.Managed, _
                                              Filter.Linear, Filter.Linear, 0)
            texCube2 = TextureLoader.FromFile(D3DDev, _
                                              Application.StartupPath + "\cube2.bmp", _
                                              256, 256, 1, 0, _
                                              Format.X8R8G8B8, Pool.Managed, _
                                              Filter.Linear, Filter.Linear, 0)

        ElseIf D3DRoot.CheckDeviceFormat(0, _
                                         DeviceType.Hardware, _
                                         AdapterFmt, _
                                         0, _
                                         ResourceType.Textures, _
                                         Format.R5G6B5 _
                                         ) Then
            texCube1 = TextureLoader.FromFile(D3DDev, _
                                              Application.StartupPath + "\cube1.bmp", _
                                              256, 256, 1, 0, _
                                              Format.R5G6B5, Pool.Managed, _
                                              Filter.Linear, Filter.Linear, 0)
            texCube2 = TextureLoader.FromFile(D3DDev, _
                                              Application.StartupPath + "\cube2.bmp", _
                                              256, 256, 1, 0, _
                                              Format.R5G6B5, Pool.Managed, _
                                              Filter.Linear, Filter.Linear, 0)

        Else
            Throw New Exception("cube textures cannot be created: no support for 16 bit (R5G6B5) or 32bit (X8R8G8B8) textures!")
        End If
    End Sub


    '[--------------------------------------------------------------]
    '   loadGeometry()
    '   this will create and/or load any geometry that we're
    '   going to be using throughout the application.
    '[--------------------------------------------------------------]
    Private Sub loadGeometry()
        Try
            'in this sample, we dont actually LOAD geometry from a file,
            'rather we LOAD it into memory.
            '[------------------------------------]
            '   LOAD 2D GEOMETRY INTO VBUFFERS
            '[------------------------------------]
            vbMenu = New VertexBuffer(GetType(CustomVertex.TransformedTextured), _
                                      4, D3DDev, 0, CustomVertex.TransformedTextured.Format, _
                                      Pool.Default)

            Dim v() As CustomVertex.TransformedTextured = CType(vbMenu.Lock(0, 0), _
                                                          CustomVertex.TransformedTextured())

            Const iMBWidth As Integer = 158
            Const iMBHeight As Integer = 93

            'use the constructors to fill the data.
            'laid out such that first line = coordinate (x,y,z,rhw) 
            'and second line = texcoord (u,v)
            v(0) = New CustomVertex.TransformedTextured(0, 86, 0, 1, _
                                                        0, 2 / 127)
            v(1) = New CustomVertex.TransformedTextured(iMBWidth, 86, 0, 1, _
                                                        iMBWidth / 255, 2 / 127)
            v(2) = New CustomVertex.TransformedTextured(0, 86 + iMBHeight, 0, 1, _
                                                        0, iMBHeight / 127)
            v(3) = New CustomVertex.TransformedTextured(iMBWidth, 86 + iMBHeight, 0, 1, _
                                                        iMBWidth / 255, iMBHeight / 127)

            vbMenu.Unlock()

            '[------------------------------------]
            '   LOAD 3D GEOMETRY INTO VBUFFERS
            '[------------------------------------]
            vbCube = New VertexBuffer(GetType(CustomVertex.PositionTextured), _
                                      36, D3DDev, 0, CustomVertex.PositionTextured.Format, _
                                      Pool.Managed)

            'use CType() to typecast a pointer to the 
            'memory into an array of position&textured vertices
            Dim vCube As CustomVertex.PositionTextured() = CType(vbCube.Lock(0, 0), _
                                                          CustomVertex.PositionTextured())

            '2a. Generate top plane
            vCube(0) = New CustomVertex.PositionTextured(-0.5, 0.5, 0.5, 0, 0)
            vCube(1) = New CustomVertex.PositionTextured(0.5, 0.5, 0.5, 1, 0)
            vCube(2) = New CustomVertex.PositionTextured(-0.5, 0.5, -0.5, 0, 1)

            vCube(3) = New CustomVertex.PositionTextured(0.5, 0.5, -0.5, 1, 1)
            vCube(4) = New CustomVertex.PositionTextured(-0.5, 0.5, -0.5, 0, 1)
            vCube(5) = New CustomVertex.PositionTextured(0.5, 0.5, 0.5, 1, 0)

            '2b. Generate bottom plane
            vCube(6) = New CustomVertex.PositionTextured(-0.5, -0.5, -0.5, 1, 0)
            vCube(7) = New CustomVertex.PositionTextured(0.5, -0.5, 0.5, 0, 1)
            vCube(8) = New CustomVertex.PositionTextured(-0.5, -0.5, 0.5, 0, 0)

            vCube(9) = New CustomVertex.PositionTextured(0.5, -0.5, 0.5, 0, 1)
            vCube(10) = New CustomVertex.PositionTextured(-0.5, -0.5, -0.5, 1, 0)
            vCube(11) = New CustomVertex.PositionTextured(0.5, -0.5, -0.5, 1, 1)

            '2c. Generate 'left' plane
            vCube(12) = New CustomVertex.PositionTextured(-0.5, 0.5, -0.5, 0, 0)
            vCube(13) = New CustomVertex.PositionTextured(-0.5, -0.5, -0.5, 1, 0)
            vCube(14) = New CustomVertex.PositionTextured(-0.5, 0.5, 0.5, 0, 1)

            vCube(15) = New CustomVertex.PositionTextured(-0.5, 0.5, 0.5, 0, 1)
            vCube(16) = New CustomVertex.PositionTextured(-0.5, -0.5, -0.5, 1, 0)
            vCube(17) = New CustomVertex.PositionTextured(-0.5, -0.5, 0.5, 1, 1)

            '2d. Generate 'right' plane
            vCube(18) = New CustomVertex.PositionTextured(0.5, 0.5, -0.5, 0, 0)
            vCube(19) = New CustomVertex.PositionTextured(0.5, 0.5, 0.5, 1, 0)
            vCube(20) = New CustomVertex.PositionTextured(0.5, -0.5, -0.5, 0, 1)

            vCube(21) = New CustomVertex.PositionTextured(0.5, -0.5, 0.5, 1, 1)
            vCube(22) = New CustomVertex.PositionTextured(0.5, -0.5, -0.5, 0, 1)
            vCube(23) = New CustomVertex.PositionTextured(0.5, 0.5, 0.5, 1, 0)

            '2e. Generate 'Back' plane
            vCube(24) = New CustomVertex.PositionTextured(-0.5, 0.5, -0.5, 0, 0)
            vCube(25) = New CustomVertex.PositionTextured(0.5, 0.5, -0.5, 1, 0)
            vCube(26) = New CustomVertex.PositionTextured(-0.5, -0.5, -0.5, 0, 1)

            vCube(27) = New CustomVertex.PositionTextured(0.5, -0.5, -0.5, 1, 1)
            vCube(28) = New CustomVertex.PositionTextured(-0.5, -0.5, -0.5, 0, 1)
            vCube(29) = New CustomVertex.PositionTextured(0.5, 0.5, -0.5, 1, 0)

            '2f. Generate 'Front' plane
            vCube(30) = New CustomVertex.PositionTextured(0.5, 0.5, 0.5, 1, 0)
            vCube(31) = New CustomVertex.PositionTextured(-0.5, 0.5, 0.5, 0, 0)
            vCube(32) = New CustomVertex.PositionTextured(-0.5, -0.5, 0.5, 0, 1)

            vCube(33) = New CustomVertex.PositionTextured(-0.5, -0.5, 0.5, 0, 1)
            vCube(34) = New CustomVertex.PositionTextured(0.5, -0.5, 0.5, 1, 1)
            vCube(35) = New CustomVertex.PositionTextured(0.5, 0.5, 0.5, 1, 0)

            vbCube.Unlock()


        Catch err As Exception
            MsgBox("loadGeometry(): " + Chr(13) + Chr(13) + err.ToString())
            Throw New Exception("could not complete loadGeometry()")
        End Try
    End Sub


    '[--------------------------------------------------------------]
    '   wireframe()
    '   allows the host to specify if we want
    '   to render in wireframe or not.
    '[--------------------------------------------------------------]
    Public Property wireframe() As Boolean
        Get
            Return bRenderWireframe
        End Get
        Set(ByVal Value As Boolean)
            'error check, then change the state.
            If Not bInitOkay Then Throw _
                New Exception("Unable to change to wireframe rendering: engine was not properly initialised")

            bRenderWireframe = Value

            If Value Then
                D3DDev.RenderState.FillMode = FillMode.WireFrame
            Else
                D3DDev.RenderState.FillMode = FillMode.Solid
            End If
        End Set
    End Property


    '[--------------------------------------------------------------]
    '   useTextures()
    '   allows the host to tell the engine
    '   to render with/without textures applied
    '   to the surfaces.
    '[--------------------------------------------------------------]
    Public Property useTextures() As Boolean
        Get
            Return bRenderTextures
        End Get
        Set(ByVal Value As Boolean)
            'error check, then change the state.
            If Not bInitOkay Then Throw _
                New Exception("Unable to change texture rendering: engine was not properly initialised")

            bRenderTextures = Value
        End Set
    End Property


    '[--------------------------------------------------------------]
    '   backFaceCulling()
    '   turns back face culling on or off.
    '[--------------------------------------------------------------]
    Public Property backFaceCulling() As Boolean
        Get
            Return bBackFaceCulling
        End Get
        Set(ByVal Value As Boolean)
            'error check, then change the state.
            If Not bInitOkay Then Throw _
                New Exception("Unable to change back face culling status: engine was not properly initialised")

            bBackFaceCulling = Value

            If Value Then
                D3DDev.RenderState.CullMode = Cull.CounterClockwise
            Else
                D3DDev.RenderState.CullMode = Cull.None
            End If
        End Set
    End Property

    '[--------------------------------------------------------------]
    '   drawFrameRate()
    '   do we draw the frame rate on the screen?
    '[--------------------------------------------------------------]
    Public Property drawFrameRate() As Boolean
        Get
            Return bDrawFPS
        End Get
        Set(ByVal Value As Boolean)
            'error check, then change the state.
            If Not bInitOkay Then Throw _
                New Exception("Unable to change frame-rate rendering: engine was not properly initialised")

            bDrawFPS = Value
        End Set
    End Property

End Class
